"""
Description: Stand-alone script that applies a machine learning model to classify wildfire retardant from Sentinel-2 imagery.
Outputs: Accuracy metrics and classification images on a per site basis.
Date: 20230104
Author: Troy Saltiel
Email: troy.saltiel@pnnl.gov
Credits: Some of the code was adapted from Chris Holden, https://github.com/ceholden/open-geo-tutorial
"""

from osgeo import gdal, ogr, gdal_array
import os
import pickle
import numpy as np
import pandas as pd
import sklearn

# File directory structure:
# difference
# input
#   - train
#     - site 1
#       - site_postfire.tif
#       - site_prefire.tif
#       - site_truth.shp
#     - site 2
#     - etc
#   - test
#     - site 1
#     - etc
# model
# prediction
#   - accuracy
#   - class
#   - overall
# truth

########## USER DEFINED VALUES ##############

# Script parameter options should match trained model
classification = 'binary' # binary or multiclass
imagery = 'difference' # difference or postfire
task = 'retardant' # retardant or burned - note, only retardant works right now, only matters for binary classification
class_attribute = 'ATTRIBUTE=' + 'Classvalue'  # Modify second string according to class value name in shapefile
no_data = 0  # Set no data value for training shapefile/raster
no_data_train_img = [no_data]  # Storing no data value in a list for GDAL in shapefile rasterization step
bands = [2, 3, 4, 5, 8, 11, 12]  # Bands selected from imagery, should be same as in input model
model_name = 'retardant_gb_081720210936.sav' # The trained model from model_training.py

working_dir = os.getcwd()
train_dir = os.path.join(working_dir, 'input', 'train')
test_dir = os.path.join(working_dir, 'input', 'test')

#############################################

def get_class_names():
    """
    Summary:
    Create a list of class names used for class accuracy metrics.

    Parameters:
    None - uses two global variables, classification and imagery

    Returns:
    class_names (list) - a list of strings representing class names
    """
    # binary, retardant or burned, difference or postfire
    if classification == 'binary':
        class_names = [task, "background"]

    # multiclass & postfire
    elif classification == 'multiclass' and imagery == 'postfire':
        class_names = ["retardant", "burned", "unburned", "cloud postfire", "shadow postfire"]

    # multiclass & difference
    elif classification == 'multiclass' and imagery == 'difference':
        class_names = ['retardant', 'burned', 'unburned', 'cloud postfire', 'cloud prefire', 'shadow postfire', 'shadow prefire']

    else:
        raise ValueError('Please select either binary or multiclass classification and postfire or difference imagery.')

    return class_names


def set_classes(truth):
    """
    Summary:
    Reclassifies or removes classes depending on script parameters

    Parameters:
    truth (array) - the original array of reference data

    Returns:
    select_classes (array) - array of reference data altered according to script parameters
    """
    # binary, retardant, difference or postfire
    if classification == 'binary' and task == 'retardant':
        select_classes = np.where(truth < 2, truth, 2)

    # binary, burned, difference or postfire
    elif classification == 'binary' and task == 'burned':
        select_classes = np.where(truth < 3, truth, 3)
        select_classes = np.where(select_classes != 1, select_classes, 3)

    # multiclass & postfire
    elif classification == 'multiclass' and imagery == 'postfire':
        select_classes = np.where(truth != 5, truth, 0)
        select_classes = np.where(select_classes != 7, select_classes, 0)

    # multiclass & difference
    else:
        select_classes = truth

    return select_classes


def file_list(path):
    """
    Summary:
    Reads all of the .tif and .shp files in the directory.

    Parameters:
    path (str) - the path to the directory

    Returns:
    data (list) - a list of file paths
    """
    data = []

    for root, dirs, files in os.walk(path):
        for name in dirs:
            path = os.path.join(root, name)
            for file in os.listdir(path):
                if file.endswith('.tif'):
                    data.append(os.path.join(path, file))
                elif file.endswith('.shp'):
                    data.append(os.path.join(path, file))

    return data


def combine_data(train_data, test_data):
    """
    Summary:
    Combines the list of data paths for training and testing data, then sorts alphabetically.

    Parameters:
    train_data (list) - a list of paths to training data
    test_data (list) - a list of paths to testing data

    Returns:
    data (list) - a list of all paths sorted alphabetically
    """
    data = []

    for file in train_data:
        data.append(file)

    for file in test_data:
        data.append(file)

    data = sorted(data)

    return data


def file_prep(data):
    """
    Summary:
    Organizes files before processing begins.

    Parameters:
    data (list) - a list of the input data files

    Returns:
    site (list) - the site names
    postfire (list) - file paths to the postfire images
    prefire (list) - file paths to the prefire images
    truth (list) - file paths to the truth shapefiles
    model_path (str) - the full path to the model file
    """

    site = []
    postfire = []
    prefire = []
    truth = []

    for file in data[0::3]:
        n = os.path.basename(file)
        site.append(n[:-13])
    for file in data[0::3]:
        postfire.append(file)
    for file in data[1::3]:
        prefire.append(file)
    for file in data[2::3]:
        truth.append(file)

    # Get model path
    model_path = os.path.join(working_dir, 'model', model_name)

    return site, postfire, prefire, truth, model_path


def img_check(postfire, prefire, site):
    """
    Summary: Checks basic image properties.

    Parameters:
    postfire (str) - the path to the postFire image
    prefire (str) - the path to the preFire image
    site (str) - the site name

    Returns:
    postfire_ncol (int) - the number of columns in the postfire image
    postfire_nrow (int) - the number of rows in the postfire image
    postfire_proj (str) - the projection of the postfire image
    postfire_ext (str) - the extent of the postfire image
    postfire_dtype (str) - the data type of the postfire image
    """
    # Tell GDAL to throw Python exceptions, and register all drivers
    gdal.UseExceptions()
    gdal.AllRegister()

    # Open the images
    postfire_ds = gdal.Open(postfire, gdal.GA_ReadOnly)
    prefire_ds = gdal.Open(prefire, gdal.GA_ReadOnly)

    # Get image info
    postfire_ncol = postfire_ds.RasterXSize
    postfire_nrow = postfire_ds.RasterYSize
    postfire_bands = postfire_ds.RasterCount
    postfire_proj = postfire_ds.GetProjectionRef()
    postfire_ext = postfire_ds.GetGeoTransform()
    postfire_dtype = gdal_array.GDALTypeCodeToNumericTypeCode(postfire_ds.GetRasterBand(1).DataType)

    prefire_bands = postfire_ds.RasterCount
    prefire_proj = prefire_ds.GetProjectionRef()
    prefire_ext = prefire_ds.GetGeoTransform()
    prefire_dtype = gdal_array.GDALTypeCodeToNumericTypeCode(prefire_ds.GetRasterBand(1).DataType)

    # Check raster size
    if postfire_ncol > 10000 or postfire_nrow > 10000:
        print('At least one dimension of the imagery contains more than 10000 px. Please be patient for processing.')

    # Check number of bands
    if postfire_bands != prefire_bands:
        print('Postfire image bands: ' + str(postfire_bands))
        print('Prefire image bands: ' + str(prefire_bands))
        raise ValueError('Image datasets do not have the same number of bands.')

    # Check projections
    if postfire_proj != prefire_proj:
        print('Postfire and prefire image projections do not match.')
        print('Postfire image projection: ' + str(postfire_proj))
        print('Prefire image projection: ' + str(prefire_proj))

    # Check image extent
    if postfire_ext != prefire_ext:
        print('Image extents do not match.')
        print('Postfire image extent: ' + postfire_ext)
        print('Prefire image extent: ' + prefire_ext)

    # Check data types
    if postfire_dtype != prefire_dtype:
        raise ValueError('Image data types do not match!')

    print('Completed image checks for site: ' + site)

    return postfire_ncol, postfire_nrow, postfire_proj, postfire_ext, postfire_dtype


# This function should not be necessary in the final code
def extract_bands(postfire, prefire, bands):
    """
    Summary:
    Reads the post- and pre- fire images and extracts the desired bands.

    Parameters:
    postfire (str) - the file path to the postfire image
    prefire (str) - the file path to the prefire image
    bands (list) - a list of the desired bands to be extracted

    Returns:
    postfire_img (array) - the post fire image with desired bands extracted
    prefire_img (array) - the pre fire image with the desired bands extracted
    """
    # Initialize the pre- and post- fire image data
    prefire_ds = gdal.Open(prefire, gdal.GA_ReadOnly)
    prefire_img = np.zeros((prefire_ds.RasterYSize, prefire_ds.RasterXSize, len(bands)),
                           gdal_array.GDALTypeCodeToNumericTypeCode(prefire_ds.GetRasterBand(1).DataType))
    postfire_ds = gdal.Open(postfire, gdal.GA_ReadOnly)
    postfire_img = np.zeros((postfire_ds.RasterYSize, postfire_ds.RasterXSize, len(bands)),
                            gdal_array.GDALTypeCodeToNumericTypeCode(postfire_ds.GetRasterBand(1).DataType))

    # Tighten the band index from the original indices to the new number of bands
    idx = []
    for i in range(len(bands)):
        idx.append(i)

    # Read the bands from the source data set into a new data set
    zip_obj = zip(idx, bands)
    for i, j in zip_obj:
        prefire_img[:, :, i] = prefire_ds.GetRasterBand(j).ReadAsArray()
        postfire_img[:, :, i] = postfire_ds.GetRasterBand(j).ReadAsArray()

    print('Bands successfully extracted.')

    return postfire_img, prefire_img


def img_diff(postfire_img, prefire_img):
    """
    Summary:
    Subtracts the prefire_img array from the postfire_img array

    Parameters:
    postfire_img (array) - the post fire image with desired bands extracted
    prefire_img (array) - the pre fire image with the desired bands extracted

    Returns:
    diff (array) - image difference array
    """
    # Subtract prefire image from postfire image
    diff = np.subtract(postfire_img, prefire_img)

    print('Completed image differencing. ')

    return diff


def validation_exists(site, truth):
    """
    Summary:
    Checks for a validation shapefile at the site.

    Parameters:
    site (str) - the site name
    truth (list) - a list of all truth shapefiles

    Returns:
    truth_file (str) - the file path for the validation truth file
    """
    truth_file = None

    for i in truth:
        base_name = os.path.basename(i)
        if site == base_name[:-10]:
            truth_file = i

    return truth_file


def rasterize_polys(site, postfire, truth):
    """
    Summary:
    Rasterizes the truth polygons.

    Parameters:
    site (str) - the site name
    postfire (str) - path to the postfire image; will be used for image information
    truth (str) - path to the shapefile

    Returns:
    truth_image - the file name for the truth image
    """
    # Open the shapefile
    dataset = ogr.Open(truth)

    # Print an error if the dataset does not exist
    if not dataset:
        print('Error: could not open dataset')

    # Get layer from shapefile
    layer = dataset.GetLayerByIndex(0)

    # Open raster image
    raster_ds = gdal.Open(postfire, gdal.GA_ReadOnly)

    # Set file name to save the rasterized training image
    truth_image = site + '_truth.tif'

    # Join full directory to file name
    working_dir = os.getcwd()
    truth_image = os.path.join(working_dir, 'truth', truth_image)

    # Fetch number of rows and columns
    ncol = raster_ds.RasterXSize
    nrow = raster_ds.RasterYSize

    # Fetch projection and extent
    proj = raster_ds.GetProjectionRef()
    ext = raster_ds.GetGeoTransform()

    raster_ds = None

    # Create the raster dataset
    memory_driver = gdal.GetDriverByName('GTiff')
    out_raster_ds = memory_driver.Create(truth_image, ncol, nrow, 1, gdal.GDT_Byte)

    # Set the ROI image's projection and extent to our input raster's projection and extent
    out_raster_ds.SetProjection(proj)
    out_raster_ds.SetGeoTransform(ext)

    # Fill our output band with the 0 blank, no class label, value
    b = out_raster_ds.GetRasterBand(1)
    b.Fill(0)

    # Rasterize the shapefile layer to our new dataset
    status = gdal.RasterizeLayer(out_raster_ds,  # output to our new dataset
                                 [1],  # output to our new dataset's first band
                                 layer,  # rasterize this layer
                                 None, None,  # don't worry about transformations since we're in same projection
                                 no_data_train_img,  # burn value 0
                                 ['ALL_TOUCHED=TRUE',  # rasterize all pixels touched by polygons
                                  class_attribute]  # put raster values according to the 'id' field values
                                 )

    # Close dataset
    out_raster_ds = None

    if status != 0:
        print("Truth polygon rasterization failed.")
    else:
        print("Truth polygon rasterization successful.")

    return truth_image


def create_validation_data(img, truth_image):
    """
    Summary:
    Creates image and label datasets for prediction validation.

    Parameters:
    img (array) - the image array
    truth_image (str) - path to the truth image file name

    Returns:
    X (array) - the image validation dataset
    y (array) - the label validation dataset
    """
    # Tell GDAL to throw Python exceptions, and register all drivers
    gdal.UseExceptions()
    gdal.AllRegister()

    # Read in our image and ROI image
    roi_ds = gdal.Open(truth_image, gdal.GA_ReadOnly)
    roi = roi_ds.GetRasterBand(1).ReadAsArray().astype(np.uint8)
    roi = set_classes(roi)

    # Print number of pixels in each class
    classes = np.unique(roi)
    for c in classes:
        print('Class {c} contains {n} pixels'.format(c=c,
                                                     n=(roi == c).sum()))
    # Create X (image data) and y (label data) variables from training image for model training/testing
    X = img[roi != no_data, :]
    y = roi[roi != no_data]

    return X, y


def get_accuracy(X, y, site, model_path):
    """
    Summary:
    Tests the model with the validation data and reports overall accuracy and class accuracy metrics.

    Parameters:
    X (array) - the image data
    y (array) - the label data
    site (str) - the site name
    model_path (str) - the full model path

    Returns:
    None - saves confusion matrix and class accuracy metrics to disk
    """
    classes = get_class_names()

    # Load a saved model
    model = pickle.load(open(model_path, 'rb'))

    # Report model score on test dataset
    print('Testing model overall accuracy. ')
    print(site + ' model test overall accuracy: ' + str(model.score(X, y)))

    # Create file names for accuracy metrics
    out_matrix = site + '_' + model_name[:-4] + '_matrix' + '.csv'
    out_matrix = os.path.join(working_dir, 'prediction', 'accuracy', out_matrix)
    out_acc = site + '_' + model_name[:-4] + '_class_accuracy' + '.csv'
    out_acc = os.path.join(working_dir, 'prediction', 'accuracy', out_acc)

    # Set up a dataframe for the confusion matrix
    print('Predicting and creating confusion matrix. ')
    df = pd.DataFrame()
    df['truth'] = y
    df['predict'] = model.predict(X)

    # Cross-tabulate predictions
    matrix = pd.crosstab(df['predict'], df['truth'], margins=True)

    # Calculate precision, recall, and F1 score for each class, then store in a data frame
    try:
        accuracy = pd.DataFrame(columns=['Classvalue', 'Name', 'Precision', 'Recall'])
        class_value = []
        recall = []
        precision = []

        for column in matrix.columns[:-1]:
            class_value.append(column)
            recall.append((matrix[column][column]) / (matrix[column]['All']))
            precision.append((matrix[column][column]) / (matrix['All'][column]))

        accuracy['Classvalue'] = class_value
        accuracy['Name'] = classes
        accuracy['Precision'] = precision
        accuracy['Recall'] = recall
        accuracy['F1'] = (
                    2 * ((accuracy['Precision'] * accuracy['Recall']) / (accuracy['Precision'] + accuracy['Recall'])))

        # Save CSV files to current directory for confusion matrix and class accuracy data frame
        print('Saving confusion matrix and class-based accuracy metrics to file.')
        matrix.to_csv(out_matrix)
        accuracy.to_csv(out_acc)
    except (KeyError, ValueError):
        print(
            'Failed to produce class-based accuracy metrics. Ensure class names are correct and check if the model '
            'predicted for all classes.')
        print('Saving confusion matrix to file.')
        matrix.to_csv(out_matrix)


def classify(img, ncol, nrow, proj, ext, model_path, site):
    """
    Summary:
    Classifies an image array with a trained model, then writes the prediction to a .tif image.

    Parameters:
    img (array) - the image array for prediction
    ncol (int) - the number of rows in the image
    nrow (int) - the number of columns in the image
    proj (str) - the projection of the image
    ext (str) - the extent of the image
    model_path (str) - the path to the model file
    site (str) - the site name

    Returns:
    None - saves a classification raster and class probability rasters to disk
    """
    # Tell GDAL to throw Python exceptions, and register all drivers
    gdal.UseExceptions()
    gdal.AllRegister()

    # Create prediction file name
    out_img = os.path.join(working_dir, 'prediction', 'overall', site)
    out_img = out_img + '_prediction_' + model_name[:-4] + '.tif'

    # Load a saved model
    model = pickle.load(open(model_path, 'rb'))

    new_shape = (img.shape[0] * img.shape[1], img.shape[2])
    img = img[:, :, :int(img.shape[2])].reshape(new_shape)
    img = np.nan_to_num(img)

    # Predict for each pixel
    print('Predicting on difference image.')
    class_prediction = model.predict(img)
    print('Prediction completed.')

    # Re-shape the prediction to the original extent of the image
    class_prediction = class_prediction.reshape((nrow, ncol))
    print('Reshaped back to {}'.format(class_prediction.shape))

    # Note that the projection, extent, cols, and rows were defined from the input imagery earlier in the script
    out_raster_ds = None
    memory_driver = gdal.GetDriverByName('GTiff')
    out_raster_ds = memory_driver.Create(out_img, ncol, nrow, 1, gdal.GDT_Byte)

    # Set the prediction image projection and extent to our input raster's projection and extent
    out_raster_ds.SetProjection(proj)
    out_raster_ds.SetGeoTransform(ext)

    # Fill our output band with the 0 blank, no class label, value
    out_raster_ds.GetRasterBand(1).WriteArray(class_prediction)
    out_raster_ds.FlushCache()
    print('Saved classification raster to file.')

#     # Predict probabilities for each class and save a raster for each
#     working_dir = os.getcwd()
#     class_name = os.path.join(working_dir, 'prediction', 'class', site)
#     class_name = class_name + 'prediction_' + model_name[:-4] + '_class'

#     for c in class_values:
#         print('Predicting for class {}'.format(str(c)))
#         class_file = class_name + str(c) + '.tif'

#         idx = c-1
#         prediction = model.predict_proba(diff)[:,idx]
#         prediction = prediction.reshape((nrow, ncol))

#         out_raster_ds = None
#         memory_driver = gdal.GetDriverByName('GTiff')
#         out_raster_ds = memory_driver.Create(class_file, ncol, nrow, 1, gdal.GDT_Float32)

#         out_raster_ds.SetProjection(proj)
#         out_raster_ds.SetGeoTransform(ext)

#         out_raster_ds.GetRasterBand(1).WriteArray(prediction)
#         out_raster_ds.FlushCache()
#         print('Saved probability raster to file.')


def main():
    train_data = file_list(train_dir)
    test_data = file_list(test_dir)
    data = combine_data(train_data, test_data)
    site, postfire, prefire, truth, model_path = file_prep(data)

    for site, postfire, prefire in zip(site, postfire, prefire):
        print('Beginning processing for : ' + site)
        ncol, nrow, proj, ext, dtype = img_check(postfire, prefire, site)
        postfire_img, prefire_img = extract_bands(postfire, prefire, bands)
        diff = img_diff(postfire_img, prefire_img)
        truth_file = validation_exists(site, truth)
        if not truth_file:
            print('No validation data exists for ' + site)
        else:
            truth_image = rasterize_polys(site, postfire, truth_file)
            if imagery == 'postfire':
                X, y = create_validation_data(postfire_img, truth_image)
                get_accuracy(X, y, site, model_path)
            else:
                X, y = create_validation_data(diff, truth_image)
                get_accuracy(X, y, site, model_path)
        if imagery == 'postfire':
            classify(postfire_img, ncol, nrow, proj, ext, model_path, site)
        else:
            classify(diff, ncol, nrow, proj, ext, model_path, site)
        print('Completed processing for: ' + site)


if __name__ == "__main__":
    main()
