"""
Description: Stand-alone script that trains a machine learning model to classify wildfire retardant from Sentinel-2 imagery.
Outputs: Model file, model metadata, training and testing accuracy.
Date: 20230104
Author: Troy Saltiel
Email: troy.saltiel@pnnl.gov
Credits: Some of the code was adapted from Chris Holden, https://github.com/ceholden/open-geo-tutorial
"""

from osgeo import gdal, ogr, gdal_array

import datetime
import os
import pickle
import numpy as np
import pandas as pd

from sklearn.model_selection import GridSearchCV, LeaveOneGroupOut
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.svm import SVC
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

# File directory structure:
# difference
# input
#   - train
#     - site 1
#       - site_postfire.tif
#       - site_prefire.tif
#       - site_truth.shp
#     - site 2
#     - etc
#   - test
#     - site 1
#     - etc
# model
# prediction
#   - accuracy
#   - class
#   - overall
# truth

########## USER DEFINED VALUES ##############

model_name = 'gb'  # svm, rf, or gb
classification = 'multiclass'  # binary or multiclass
class_weight = 'balanced'  # None, 'balanced', 'balanced subsample', or dict
imagery = 'difference'  # difference or postfire
task = 'retardant'  # retardant or burned - note, only retardant works right now, only matters for binary classification
no_data = 0  # set to no data value of class schema
class_value = 'ATTRIBUTE=' + 'Classvalue'  # change second string to attribute name for class values
bands = [2, 3, 4, 5, 8, 11, 12]  # select bands from values from 1 through 12
working_dir = os.getcwd()
train_dir = os.path.join(working_dir, 'input', 'train')
test_dir = os.path.join(working_dir, 'input', 'test')

no_data_train_img = [no_data]

now = datetime.datetime.now()
date = now.strftime("%m%d%Y%H%M")

meta = ['Task: ' + task,
        'Model: ' + model_name,
        'Imagery: ' + imagery,
        'Classification: ' + classification,
        'Class weight: ' + class_weight,
        'no_data: ' + str(no_data),
        'Bands: ' + ', '.join(str(b) for b in bands)
        ]

#############################################

def get_class_names():
    """
    Summary:
    Create a list of class names used for class accuracy metrics.

    Parameters:
    None - uses two global variables, classification and imagery

    Returns:
    class_names (list) - a list of strings representing class names
    """
    # binary, retardant or burned, difference or postfire
    if classification == 'binary':
        class_names = [task, "background"]

    # multiclass & postfire
    elif classification == 'multiclass' and imagery == 'postfire':
        class_names = ["retardant", "burned", "unburned", "cloud postfire", "shadow postfire"]

    # multiclass & difference
    elif classification == 'multiclass' and imagery == 'difference':
        class_names = ['retardant', 'burned', 'unburned', 'cloud postfire', 'cloud prefire', 'shadow postfire', 'shadow prefire']

    else:
        raise ValueError('Please select either binary or multiclass classification and postfire or difference imagery.')

    return class_names


def set_classes(truth):
    """
    Summary:
    Reclassifies or removes classes depending on script parameters

    Parameters:
    truth (array) - the original array of reference data

    Returns:
    select_classes (array) - array of reference data altered according to script parameters
    """
    # binary, retardant, difference or postfire
    if classification == 'binary' and task == 'retardant':
        select_classes = np.where(truth < 2, truth, 2)

    # binary, burned, difference or postfire
    elif classification == 'binary' and task == 'burned':
        select_classes = np.where(truth < 3, truth, 3)
        select_classes = np.where(select_classes != 1, select_classes, 3)

    # multiclass & postfire
    elif classification == 'multiclass' and imagery == 'postfire':
        select_classes = np.where(truth != 5, truth, 0)
        select_classes = np.where(select_classes != 7, select_classes, 0)

    # multiclass & difference
    else:
        select_classes = truth

    return select_classes


def file_list(path):
    """
    Summary:
    Reads all of the .tif and .shp files in the directory.

    Parameters:
    path (str) - the path to the directory

    Returns:
    data (list) - a list of file paths sorted alphabetically
    """
    data = []

    for root, dirs, files in os.walk(path):
        for name in dirs:
            path = os.path.join(root, name)
            for file in os.listdir(path):
                if file.endswith('.tif'):
                    data.append(os.path.join(path, file))
                elif file.endswith('.shp'):
                    data.append(os.path.join(path, file))

    data = sorted(data)

    return data


def file_prep(data):
    """
    Summary:
    Organizes files before processing begins.

    Parameters:
    None - however, files and directories need to be organized before starting.

    Returns:
    site (list) - the site names
    postfire (list) - file paths to the postfire images
    prefire (list) - file paths to the prefire images
    truth (list) - file paths to the truth shapefiles
    """

    site = []
    postfire = []
    prefire = []
    truth = []

    for file in data[0::3]:
        n = os.path.basename(file)
        site.append(n[:-13])
    for file in data[0::3]:
        postfire.append(file)
    for file in data[1::3]:
        prefire.append(file)
    for file in data[2::3]:
        truth.append(file)

    return site, postfire, prefire, truth


def rasterize_polys(site, postfire, truth):
    """
    Summary:
    Rasterizes the truth polygons.

    Parameters:
    site (str) - the site name
    postfire (str) - path to the postfire image; will be used for image information
    truth (str) - path to the reference shapefile

    Returns:
    truth_image - the file name for the truth image
    """
    # Open the shapefile
    dataset = ogr.Open(truth)

    # Print an error if the dataset does not exist
    if not dataset:
        print('Error: could not open dataset')

    # Get layer from shapefile
    layer = dataset.GetLayerByIndex(0)

    # Open raster image
    raster_ds = gdal.Open(postfire, gdal.GA_ReadOnly)

    # Set file name to save the rasterized training image
    truth_image = site + '_truth.tif'

    # Join full directory to file name
    working_dir = os.getcwd()
    truth_image = os.path.join(working_dir, 'truth', truth_image)

    # Fetch number of rows and columns
    ncol = raster_ds.RasterXSize
    nrow = raster_ds.RasterYSize

    # Fetch projection and extent
    proj = raster_ds.GetProjectionRef()
    ext = raster_ds.GetGeoTransform()

    raster_ds = None

    # Create the raster dataset
    memory_driver = gdal.GetDriverByName('GTiff')
    out_raster_ds = memory_driver.Create(truth_image, ncol, nrow, 1, gdal.GDT_Byte)

    # Set the ROI image's projection and extent to our input raster's projection and extent
    out_raster_ds.SetProjection(proj)
    out_raster_ds.SetGeoTransform(ext)

    # Fill our output band with the 0 blank, no class label, value
    b = out_raster_ds.GetRasterBand(1)
    b.Fill(0)

    # Rasterize the shapefile layer to our new dataset
    status = gdal.RasterizeLayer(out_raster_ds,  # output to our new dataset
                                 [1],  # output to our new dataset's first band
                                 layer,  # rasterize this layer
                                 None, None,  # don't worry about transformations since we're in same projection
                                 no_data_train_img,  # burn value 0
                                 ['ALL_TOUCHED=TRUE',  # rasterize all pixels touched by polygons
                                  class_value]  # put raster values according to the 'id' field values
                                 )

    # Close dataset
    out_raster_ds = None

    if status != 0:
        print("Train polygon rasterization failed.")
    else:
        print("Train polygon rasterization successful.")

    return truth_image


def extract_bands(imagery, bands):
    """
    Summary:
    Reads an image and extracts the desired bands.

    Parameters:
    imagery (str) - the file path to the image
    bands (list) - a list of the desired bands to be extracted

    Returns:
    img (array) - the image with desired bands extracted
    """
    # Initialize the image data
    print(imagery)
    ds = gdal.Open(imagery, gdal.GA_ReadOnly)
    img = np.zeros((ds.RasterYSize, ds.RasterXSize, len(bands)),
                   gdal_array.GDALTypeCodeToNumericTypeCode(ds.GetRasterBand(1).DataType))

    # Tighten the band index from the original indices to the new number of bands
    idx = []
    for i in range(len(bands)):
        idx.append(i)

    # Read the bands from the source data set into a new data set
    zip_obj = zip(idx, bands)
    for i, j in zip_obj:
        img[:, :, i] = ds.GetRasterBand(j).ReadAsArray()

    print('Bands sucessfully extracted.')

    return img


def img_diff(site, postfire_img, prefire_img, postfire):
    """
    Summary:
    Subtracts the prefire image from the postfire image.

    Parameters:
    site (str) - the site name
    postfire_img (array) - the postfire image
    prefire_img (array) - the prefire image
    postfire (str) - the path to the postfire image

    Returns:
    diff (array) - the difference image
    """
    # Subtract prefire image from postfire image
    diff = np.subtract(postfire_img, prefire_img)

    # Set file name to save difference raster
    diff_file_name = site + '_difference.tif'

    # Join full directory to file name
    working_dir = os.getcwd()
    diff_file_name = os.path.join(working_dir, 'difference', diff_file_name)

    # Open prefire raster image to extract image info
    raster_ds = gdal.Open(postfire, gdal.GA_ReadOnly)

    # Fetch number of rows and columns
    ncol = raster_ds.RasterXSize
    nrow = raster_ds.RasterYSize

    # Fetch projection and extent
    proj = raster_ds.GetProjectionRef()
    ext = raster_ds.GetGeoTransform()

    raster_ds = None

    # Create the raster dataset
    memory_driver = gdal.GetDriverByName('GTiff')
    out_raster_ds = memory_driver.Create(diff_file_name, ncol, nrow, len(bands), gdal.GDT_UInt16)

    # Set the ROI image's projection and extent to our input raster's projection and extent
    out_raster_ds.SetProjection(proj)
    out_raster_ds.SetGeoTransform(ext)

    # Write bands to new file
    for i in range(len(bands)):
        j = i + 1
        out_raster_ds.GetRasterBand(j).WriteArray(diff[:, :, i])

    # Close dataset
    out_raster_ds = None

    print('postfire - prefire difference raster produced.')

    return diff


def train_data_prep_init(imagery, truth_image, group):
    """
    Summary:
    Splits image and label data.
    Initializes the training data with the first site's data.

    Parameters:
    imagery (str) - path to the image file name
    truth_image (str) - path to the truth image file name
    group (int) - the site group id value

    Returns:
    X (array) - the image dataset
    y (array) - the label dataset
    g (array) - an array filled with the site group value
    """
    # Tell GDAL to throw Python exceptions, and register all drivers
    gdal.UseExceptions()
    gdal.AllRegister()

    # Read in the ROI image
    truth_ds = gdal.Open(truth_image, gdal.GA_ReadOnly)

    truth = truth_ds.GetRasterBand(1).ReadAsArray().astype(np.uint8)
    truth = set_classes(truth)

    # Print number of pixels in each class
    truth_classes = np.unique(truth)
    for c in truth_classes:
        print('Truth class {c} contains {n} pixels'.format(c=c,
                                                           n=(truth == c).sum()))

    # Create X (image data) and y (label data) variables from training image for model training/testing
    X = imagery[truth != no_data, :]
    y = truth[truth != no_data]
    g = np.full_like(y, fill_value=group)

    return X, y, g


def train_data_prep_append(imagery, truth_image, X, y, g, group):
    """
    Summary:
    Splits image and label data.
    Initializes the training data with the first site's data.

    Parameters:
    imagery (str) - path to the image file name
    truth_image (str) - path to the truth image file name
    X (array) - the image dataset
    y (array) - the label dataset
    g (array) - an array filled with the site group value
    group (int) - the site group id value

    Returns:
    X (array) - the image dataset
    y (array) - the label dataset
    g (array) - an array filled with the site group value
    """
    # Tell GDAL to throw Python exceptions, and register all drivers
    gdal.UseExceptions()
    gdal.AllRegister()

    # Read in our image and ROI image
    truth_ds = gdal.Open(truth_image, gdal.GA_ReadOnly)

    truth = truth_ds.GetRasterBand(1).ReadAsArray().astype(np.uint8)
    truth = set_classes(truth)

    # Print number of pixels in each class
    truth_classes = np.unique(truth)
    for c in truth_classes:
        print('Truth class {c} contains {n} pixels'.format(c=c,
                                                           n=(truth == c).sum()))

    # Create X (image data) and y (label data) variables from training image for model training/testing
    X_con = imagery[truth != no_data, :]
    y_con = truth[truth != no_data]
    g_con = np.full_like(y_con, fill_value=group)

    X = np.concatenate((X, X_con))
    y = np.concatenate((y, y_con))
    g = np.concatenate((g, g_con))

    return X, y, g


def train_model(date, model_name, X, y, g):
    """
    Summary:
    Trains a model.

    Parameters:
    date (str) - Python generated date, used for unique file names
    model_name (str) - user specified model name; either svm, rf, or gb
    X (array) - the image dataset
    y (array) - the label dataset
    g (array) - the group labels

    Returns:
    model (sklearn model) - a trained sklearn model
    model_file_name (str) - the model file name
    stats (list) - a list of stats from model training
    """
    stats = []

    classes = get_class_names()
    class_names = "Classes: " + ", ".join(classes)
    stats.append(class_names)

    logo = LeaveOneGroupOut().split(X, y, g)
    stats.append("Training cross-validation method: LeaveOneGroupOut")
    print("Cross-validation groups: " + str(len(np.unique(g))))

    if model_name == 'rf':
        param_grid = {
            'bootstrap': [True],
            'max_depth': [5, 10, 20, 40, 80, 160],
            'max_features': ['auto'],
            'min_samples_leaf': [1, 2, 4, 8, 16, 32],
            'min_samples_split': [2, 4, 8, 16, 32],
            'class_weight': ['balanced'],
            'n_estimators': [100]
        }

        print('Initializing model')
        rf = RandomForestClassifier(random_state=42)
        model = GridSearchCV(estimator=rf, param_grid=param_grid,
                             cv=logo, n_jobs=16)

        print('Fitting model')
        model = model.fit(X, y)

    elif model_name == 'svm':
        param_grid = {
            'svc__C': [0.1, 1, 10, 100, 1000, 10000],
            'svc__gamma': [1, 0.1, 0.01, 0.001, 0.0001, 'scale'],
            'svc__kernel': ['rbf']
        }

        print('Initializing model')
        svm = make_pipeline(StandardScaler(), SVC(random_state=42))
        model = GridSearchCV(estimator=svm, param_grid=param_grid,
                             cv=logo, n_jobs=-1)

        print('Fitting model')
        model.fit(X, y)

    elif model_name == 'gb':
        param_grid = {
            'learning_rate': [0.001, 0.01, 0.1, 1, 10],
            'n_estimators': [100],
            'min_samples_leaf': [1, 2, 4, 8, 16],
            'min_samples_split': [2, 4, 8, 16],
            'max_features': ['auto']
        }

        print('Initializing model')
        gb = GradientBoostingClassifier(random_state=42)
        model = GridSearchCV(estimator=gb, param_grid=param_grid,
                             cv=logo, n_jobs=16)
        # Fit our model to training data
        print('Fitting model')
        model = model.fit(X, y)

    else:
        raise ValueError('Model selection must be rf, svm, or gb. Update model_name variable.')

    print('Model successfully fit to data.')
    best_params = "Best parameters: " + str(model.best_estimator_)
    print(best_params)
    stats.append(best_params)

    best_score = "Average CV score: " + str(model.best_score_)
    print(best_score)
    stats.append(best_score)

    # Set model file name
    model_file_name = task + '_' + model_name + '_' + date + '.sav'

    # Join full directory to file name
    working_dir = os.getcwd()
    model_file_name = os.path.join(working_dir, 'model', model_file_name)

    # Save model to file
    print('Saving model file.')
    pickle.dump(model.best_estimator_, open(model_file_name, 'wb'))
    print(model_file_name)

    # Create file names for accuracy metrics
    out_matrix = task + '_' + model_name + '_train_matrix_' + date + '.csv'
    out_matrix = os.path.join(working_dir, 'model', out_matrix)
    out_acc = task + '_' + model_name + '_train_class_accuracy_' + date + '.csv'
    out_acc = os.path.join(working_dir, 'model', out_acc)

    # Set up a dataframe for the confusion matrix
    df = pd.DataFrame()
    df['truth'] = y
    df['predict'] = model.predict(X)

    # Cross-tabulate predictions
    matrix = pd.crosstab(df['predict'], df['truth'], margins=True)

    # Calculate precision, recall, and F1 score for each class, then store in a data frame
    try:
        accuracy = pd.DataFrame(columns=['class_value', 'Name', 'Precision', 'Recall'])
        class_value = []
        recall = []
        precision = []

        for column in matrix.columns[:-1]:
            class_value.append(column)
            recall.append((matrix[column][column]) / (matrix[column]['All']))
            precision.append((matrix[column][column]) / (matrix['All'][column]))

        accuracy['class_value'] = class_value
        accuracy['Name'] = classes
        accuracy['Precision'] = precision
        accuracy['Recall'] = recall
        accuracy['F1'] = (
                2 * ((accuracy['Precision'] * accuracy['Recall']) / (accuracy['Precision'] + accuracy['Recall'])))

        # Save CSV files to current directory for confusion matrix and class accuracy data frame
        print('Saving confusion matrix and class-based accuracy metrics to file.')
        matrix.to_csv(out_matrix)
        accuracy.to_csv(out_acc)
    except (KeyError, ValueError):
        print(
            'Failed to produce class-based accuracy metrics. Ensure class names are correct and check if the model '
            'predicted for all classes.')
        print('Saving confusion matrix to file.')
        matrix.to_csv(out_matrix)

    return model, model_file_name, stats


def test_model(model, stats, X, y, date):
    """
    Summary:
    Tests the model on unseen data.

    Parameters:
    model (sklearn model) - a trained sklearn model
    stats (list) - a list of model stats for metadata
    X (array) - the image dataset
    y (array) - the label dataset
    date (str) - Python generated date, used for unique file names

    Returns:
    stats (list) - a list of model stats for metadata
    """
    classes = get_class_names()

    # Report model score on test dataset
    test_acc = 'Test overall accuracy: ' + str(model.score(X, y))
    stats.append(test_acc)
    print(test_acc)

    # Create file names for accuracy metrics
    out_matrix = task + '_' + model_name + '_test_matrix_' + date + '.csv'
    out_matrix = os.path.join(working_dir, 'model', out_matrix)
    out_acc = task + '_' + model_name + '_test_class_accuracy_' + date + '.csv'
    out_acc = os.path.join(working_dir, 'model', out_acc)

    # Set up a dataframe for the confusion matrix
    df = pd.DataFrame()
    df['truth'] = y
    df['predict'] = model.predict(X)

    # Cross-tabulate predictions
    matrix = pd.crosstab(df['predict'], df['truth'], margins=True)

    # Calculate precision, recall, and F1 score for each class, then store in a data frame
    try:
        accuracy = pd.DataFrame(columns=['class_value', 'Name', 'Precision', 'Recall'])
        class_value = []
        recall = []
        precision = []

        for column in matrix.columns[:-1]:
            class_value.append(column)
            recall.append((matrix[column][column]) / (matrix[column]['All']))
            precision.append((matrix[column][column]) / (matrix['All'][column]))

        accuracy['class_value'] = class_value
        accuracy['Name'] = classes
        accuracy['Precision'] = precision
        accuracy['Recall'] = recall
        accuracy['F1'] = (
                2 * ((accuracy['Precision'] * accuracy['Recall']) / (accuracy['Precision'] + accuracy['Recall'])))

        # Save CSV files to current directory for confusion matrix and class accuracy data frame
        print('Saving confusion matrix and class-based accuracy metrics to file.')
        matrix.to_csv(out_matrix)
        accuracy.to_csv(out_acc)
    except (KeyError, ValueError):
        print(
            'Failed to produce class-based accuracy metrics. Ensure class names are correct and check if the model '
            'predicted for all classes.')
        print('Saving confusion matrix to file.')
        matrix.to_csv(out_matrix)

    return stats


def training(train_dir):
    """
    Summary:
    Processing image and label data and trains a model.

    Parameters:
    train_dir (str) - path to the training data directory

    Returns:
    model (sklearn model) - a trained sklearn model
    model_file_name (str) - the model file name
    stats (list) - a list of stats from model training
    """
    group = 0
    train_data = file_list(train_dir)
    site, postfire, prefire, truth = file_prep(train_data)
    truth_image = rasterize_polys(site[0], postfire[0], truth[0])
    postfire_img = extract_bands(postfire[0], bands)
    if imagery == "difference":
        prefire_img = extract_bands(prefire[0], bands)
        diff = img_diff(site[0], postfire_img, prefire_img, postfire[0])
        X, y, g = train_data_prep_init(diff, truth_image, group)
    else:
        X, y, g = train_data_prep_init(postfire_img, truth_image, group)
    print('Site name: ' + site[0] + ' processing complete.')

    for site, postfire, prefire, truth in zip(site[1:], postfire[1:], prefire[1:], truth[1:]):
        group += 1
        truth_image = rasterize_polys(site, postfire, truth)
        postfire_img = extract_bands(postfire, bands)
        if imagery == "difference":
            prefire_img = extract_bands(prefire, bands)
            diff = img_diff(site, postfire_img, prefire_img, postfire)
            X, y, g = train_data_prep_append(diff, truth_image, X, y, g, group)
        else:
            X, y, g = train_data_prep_append(postfire_img, truth_image, X, y, g, group)
        print('Site name: ' + site + ' processing complete.')
    print('Data processing for all sites complete. Beginning training.')

    print('The X array is sized: {sz}'.format(sz=X.shape))
    print('The y array is sized: {sz}'.format(sz=y.shape))
    model, model_file_name, stats = train_model(date, model_name, X, y, g)
    print('Model training complete.')

    return model, model_file_name, stats


def testing(test_dir, model, stats):
    """
    Summary:
    Processes image and label data and tests a model.

    Parameters:
    test_dir (str) - path to the testing data directory

    Returns:
    stats (list) - a list of stats from model training and testing
    """
    group = 0
    test_data = file_list(test_dir)
    site, postfire, prefire, truth = file_prep(test_data)
    truth_image = rasterize_polys(site[0], postfire[0], truth[0])
    postfire_img = extract_bands(postfire[0], bands)

    if imagery == "difference":
        prefire_img = extract_bands(prefire[0], bands)
        diff = img_diff(site[0], postfire_img, prefire_img, postfire[0])
        X, y, g = train_data_prep_init(diff, truth_image, group)
    else:
        X, y, g = train_data_prep_init(postfire_img, truth_image, group)

    print('Site name: ' + site[0] + ' processing complete.')

    for site, postfire, prefire, truth in zip(site[1:], postfire[1:], prefire[1:], truth[1:]):
        group += 1
        truth_image = rasterize_polys(site, postfire, truth)
        postfire_img = extract_bands(postfire, bands)
        if imagery == "difference":
            prefire_img = extract_bands(prefire, bands)
            diff = img_diff(site, postfire_img, prefire_img, postfire)
            X, y, g = train_data_prep_append(diff, truth_image, X, y, g, group)
        else:
            X, y, g = train_data_prep_append(postfire_img, truth_image, X, y, g, group)
        print('Site name: ' + site + ' processing complete.')
    print('Data processing for all sites complete. Beginning testing.')
    stats = test_model(model, stats, X, y, date)
    print('Model testing complete.')

    return stats


def save_meta(model_file_name, meta, stats):
    """
    Summary:
    Saves model metadata to a text file.

    Parameters:
    model_file_name (str) - the name of the model .sav file
    meta (list) - a list of model variables to print to the text file

    Returns:
    None - Saves a .txt file to disk.
    """
    file_name = model_file_name[:-4] + '_meta.txt'
    with open(file_name, 'w') as file:
        for row in meta:
            s = "".join(map(str, row))
            file.write(s + '\n')

        for row in stats:
            s = "".join(map(str, row))
            file.write(s + '\n')

    print('Saved metadata to text file.')


def main():
    model, model_file_name, stats = training(train_dir)
    stats = testing(test_dir, model, stats)
    save_meta(model_file_name, meta, stats)


if __name__ == "__main__":
    main()
